==> bodo-2022.5.8-cp310-cp310-macosx_10_15_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    bhk__rxzb = n + 7 >> 3
    jhs__yvo = np.empty(bhk__rxzb, np.uint8)
    adfrd__csede = 0
    for tbteo__lxf in range(len(ind)):
        if ind[tbteo__lxf]:
            jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                tbteo__lxf)
            bodo.libs.int_arr_ext.set_bit_to_arr(jhs__yvo, adfrd__csede,
                jwz__aaq)
            adfrd__csede += 1
    return jhs__yvo


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    sul__kzzt = A._data[ind]
    n = len(sul__kzzt)
    jhs__yvo = get_new_null_mask_bool_index(old_mask, ind, n)
    return sul__kzzt, jhs__yvo


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    bhk__rxzb = n + 7 >> 3
    jhs__yvo = np.empty(bhk__rxzb, np.uint8)
    adfrd__csede = 0
    for tbteo__lxf in range(len(ind)):
        jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask, ind[
            tbteo__lxf])
        bodo.libs.int_arr_ext.set_bit_to_arr(jhs__yvo, adfrd__csede, jwz__aaq)
        adfrd__csede += 1
    return jhs__yvo


@register_jitable
def array_getitem_int_index(A, ind):
    fli__hhdmm = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    sul__kzzt = A._data[fli__hhdmm]
    n = len(sul__kzzt)
    jhs__yvo = get_new_null_mask_int_index(old_mask, fli__hhdmm, n)
    return sul__kzzt, jhs__yvo


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    dlsbn__hxly = numba.cpython.unicode._normalize_slice(ind, n)
    mtg__cjicg = numba.cpython.unicode._slice_span(dlsbn__hxly)
    bhk__rxzb = mtg__cjicg + 7 >> 3
    jhs__yvo = np.empty(bhk__rxzb, np.uint8)
    adfrd__csede = 0
    for tbteo__lxf in range(dlsbn__hxly.start, dlsbn__hxly.stop,
        dlsbn__hxly.step):
        jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            tbteo__lxf)
        bodo.libs.int_arr_ext.set_bit_to_arr(jhs__yvo, adfrd__csede, jwz__aaq)
        adfrd__csede += 1
    return jhs__yvo


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    sul__kzzt = np.ascontiguousarray(A._data[ind])
    jhs__yvo = get_new_null_mask_slice_index(old_mask, ind, n)
    return sul__kzzt, jhs__yvo


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for tbteo__lxf in range(n):
                A._data[idx[tbteo__lxf]] = val._data[tbteo__lxf]
                jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val.
                    _null_bitmap, tbteo__lxf)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    tbteo__lxf], jwz__aaq)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for tbteo__lxf in idx:
                A._data[tbteo__lxf] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    tbteo__lxf, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            bivn__agpi = 0
            for tbteo__lxf in range(n):
                if not bodo.libs.array_kernels.isna(idx, tbteo__lxf) and idx[
                    tbteo__lxf]:
                    A._data[tbteo__lxf] = val._data[bivn__agpi]
                    jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, bivn__agpi)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        tbteo__lxf, jwz__aaq)
                    bivn__agpi += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            bivn__agpi = 0
            for tbteo__lxf in range(n):
                if not bodo.libs.array_kernels.isna(idx, tbteo__lxf) and idx[
                    tbteo__lxf]:
                    A._data[tbteo__lxf] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        tbteo__lxf, 1)
                    bivn__agpi += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    dlsbn__hxly = numba.cpython.unicode._normalize_slice(idx, n)
    bivn__agpi = 0
    for tbteo__lxf in range(dlsbn__hxly.start, dlsbn__hxly.stop,
        dlsbn__hxly.step):
        jwz__aaq = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            bivn__agpi)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, tbteo__lxf, jwz__aaq)
        bivn__agpi += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            dlsbn__hxly = numba.cpython.unicode._normalize_slice(idx, len(A))
            for tbteo__lxf in range(dlsbn__hxly.start, dlsbn__hxly.stop,
                dlsbn__hxly.step):
                A._data[tbteo__lxf] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    tbteo__lxf, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for tbteo__lxf in idx:
                    bodo.libs.array_kernels.setna(A, tbteo__lxf)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    smi__nvnjq = bodo.libs.str_arr_ext.pre_alloc_string_array(n
                        , -1)
                    for tbteo__lxf in numba.parfors.parfor.internal_prange(n):
                        if idx[tbteo__lxf] or bodo.libs.array_kernels.isna(A,
                            tbteo__lxf):
                            smi__nvnjq[tbteo__lxf] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(smi__nvnjq,
                                tbteo__lxf)
                        else:
                            smi__nvnjq[tbteo__lxf] = A[tbteo__lxf]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        smi__nvnjq)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for tbteo__lxf in range(n):
                    if not bodo.libs.array_kernels.isna(idx, tbteo__lxf
                        ) and idx[tbteo__lxf]:
                        bodo.libs.array_kernels.setna(A, tbteo__lxf)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                dlsbn__hxly = numba.cpython.unicode._normalize_slice(idx, n)
                for tbteo__lxf in range(dlsbn__hxly.start, dlsbn__hxly.stop,
                    dlsbn__hxly.step):
                    bodo.libs.array_kernels.setna(A, tbteo__lxf)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for tbteo__lxf in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, tbteo__lxf)
                        continue
                    A[tbteo__lxf] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for tbteo__lxf in range(n):
                    if not bodo.libs.array_kernels.isna(idx, tbteo__lxf
                        ) and idx[tbteo__lxf]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, tbteo__lxf)
                            continue
                        A[tbteo__lxf] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                dlsbn__hxly = numba.cpython.unicode._normalize_slice(idx, n)
                for tbteo__lxf in range(dlsbn__hxly.start, dlsbn__hxly.stop,
                    dlsbn__hxly.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, tbteo__lxf)
                        continue
                    A[tbteo__lxf] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        tbtb__wnog = context.make_helper(builder, val_t, args[0])
        wqou__jgxi = tbtb__wnog.data
        context.nrt.incref(builder, val_t.type, wqou__jgxi)
        return wqou__jgxi
    return val_t.type(val_t), codegen

==> bodo-2022.7.1-cp310-cp310-macosx_10_15_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    cbyrk__kizza = n + 7 >> 3
    ugud__qskmr = np.empty(cbyrk__kizza, np.uint8)
    szpq__dcg = 0
    for oxv__fklbp in range(len(ind)):
        if ind[oxv__fklbp]:
            vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                oxv__fklbp)
            bodo.libs.int_arr_ext.set_bit_to_arr(ugud__qskmr, szpq__dcg,
                vevo__jhhed)
            szpq__dcg += 1
    return ugud__qskmr


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    wbxpo__uos = A._data[ind]
    n = len(wbxpo__uos)
    ugud__qskmr = get_new_null_mask_bool_index(old_mask, ind, n)
    return wbxpo__uos, ugud__qskmr


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    cbyrk__kizza = n + 7 >> 3
    ugud__qskmr = np.empty(cbyrk__kizza, np.uint8)
    szpq__dcg = 0
    for oxv__fklbp in range(len(ind)):
        vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            ind[oxv__fklbp])
        bodo.libs.int_arr_ext.set_bit_to_arr(ugud__qskmr, szpq__dcg,
            vevo__jhhed)
        szpq__dcg += 1
    return ugud__qskmr


@register_jitable
def array_getitem_int_index(A, ind):
    jdmt__ureqi = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    wbxpo__uos = A._data[jdmt__ureqi]
    n = len(wbxpo__uos)
    ugud__qskmr = get_new_null_mask_int_index(old_mask, jdmt__ureqi, n)
    return wbxpo__uos, ugud__qskmr


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    gqhik__cpaa = numba.cpython.unicode._normalize_slice(ind, n)
    zei__iifo = numba.cpython.unicode._slice_span(gqhik__cpaa)
    cbyrk__kizza = zei__iifo + 7 >> 3
    ugud__qskmr = np.empty(cbyrk__kizza, np.uint8)
    szpq__dcg = 0
    for oxv__fklbp in range(gqhik__cpaa.start, gqhik__cpaa.stop,
        gqhik__cpaa.step):
        vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            oxv__fklbp)
        bodo.libs.int_arr_ext.set_bit_to_arr(ugud__qskmr, szpq__dcg,
            vevo__jhhed)
        szpq__dcg += 1
    return ugud__qskmr


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    wbxpo__uos = np.ascontiguousarray(A._data[ind])
    ugud__qskmr = get_new_null_mask_slice_index(old_mask, ind, n)
    return wbxpo__uos, ugud__qskmr


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for oxv__fklbp in range(n):
                A._data[idx[oxv__fklbp]] = val._data[oxv__fklbp]
                vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val.
                    _null_bitmap, oxv__fklbp)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    oxv__fklbp], vevo__jhhed)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for oxv__fklbp in idx:
                A._data[oxv__fklbp] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    oxv__fklbp, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            tip__lnrng = 0
            for oxv__fklbp in range(n):
                if not bodo.libs.array_kernels.isna(idx, oxv__fklbp) and idx[
                    oxv__fklbp]:
                    A._data[oxv__fklbp] = val._data[tip__lnrng]
                    vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, tip__lnrng)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        oxv__fklbp, vevo__jhhed)
                    tip__lnrng += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            tip__lnrng = 0
            for oxv__fklbp in range(n):
                if not bodo.libs.array_kernels.isna(idx, oxv__fklbp) and idx[
                    oxv__fklbp]:
                    A._data[oxv__fklbp] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        oxv__fklbp, 1)
                    tip__lnrng += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    gqhik__cpaa = numba.cpython.unicode._normalize_slice(idx, n)
    tip__lnrng = 0
    for oxv__fklbp in range(gqhik__cpaa.start, gqhik__cpaa.stop,
        gqhik__cpaa.step):
        vevo__jhhed = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            tip__lnrng)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, oxv__fklbp,
            vevo__jhhed)
        tip__lnrng += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            gqhik__cpaa = numba.cpython.unicode._normalize_slice(idx, len(A))
            for oxv__fklbp in range(gqhik__cpaa.start, gqhik__cpaa.stop,
                gqhik__cpaa.step):
                A._data[oxv__fklbp] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    oxv__fklbp, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for oxv__fklbp in idx:
                    bodo.libs.array_kernels.setna(A, oxv__fklbp)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    aedqr__qrzn = bodo.libs.str_arr_ext.pre_alloc_string_array(
                        n, -1)
                    for oxv__fklbp in numba.parfors.parfor.internal_prange(n):
                        if idx[oxv__fklbp] or bodo.libs.array_kernels.isna(A,
                            oxv__fklbp):
                            aedqr__qrzn[oxv__fklbp] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(aedqr__qrzn,
                                oxv__fklbp)
                        else:
                            aedqr__qrzn[oxv__fklbp] = A[oxv__fklbp]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        aedqr__qrzn)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for oxv__fklbp in range(n):
                    if not bodo.libs.array_kernels.isna(idx, oxv__fklbp
                        ) and idx[oxv__fklbp]:
                        bodo.libs.array_kernels.setna(A, oxv__fklbp)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                gqhik__cpaa = numba.cpython.unicode._normalize_slice(idx, n)
                for oxv__fklbp in range(gqhik__cpaa.start, gqhik__cpaa.stop,
                    gqhik__cpaa.step):
                    bodo.libs.array_kernels.setna(A, oxv__fklbp)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for oxv__fklbp in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, oxv__fklbp)
                        continue
                    A[oxv__fklbp] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for oxv__fklbp in range(n):
                    if not bodo.libs.array_kernels.isna(idx, oxv__fklbp
                        ) and idx[oxv__fklbp]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, oxv__fklbp)
                            continue
                        A[oxv__fklbp] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                gqhik__cpaa = numba.cpython.unicode._normalize_slice(idx, n)
                for oxv__fklbp in range(gqhik__cpaa.start, gqhik__cpaa.stop,
                    gqhik__cpaa.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, oxv__fklbp)
                        continue
                    A[oxv__fklbp] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        ajbqy__mtvur = context.make_helper(builder, val_t, args[0])
        tlv__qvzc = ajbqy__mtvur.data
        context.nrt.incref(builder, val_t.type, tlv__qvzc)
        return tlv__qvzc
    return val_t.type(val_t), codegen
